"""uuid.py: Thin bifrost wrapper around python uuid (universally unique identifier) module."""

# NB: at present the code does not enforce multiprocessor safe generation of unique identifiers
# it calls uuid.uuid1() to generate new uuids, ie. using version 1 of the iso standard options

version = '25th August 2021'

# import logging
# log = logging.getLogger(__name__)

import uuid

test_mode = False
test_latest_int = 0

max_version_string_length = 10


def switch_on_test_mode(seed = 0):
   """Causes subsequent calls to new_uid() to produce integer sequence starting from successor to seed.

   arguments:
      seed (integer, default 0): The predecessor to the first uuid returned by subsequent calls to
         new_uuid()

   returns:
      None

   notes:
      call switch_off_test_mode() to reactivate normal behaviour;
      uuids generated whilst in test mode do not adhere to the iso standard;
      test mode is intended to allow replicatable behaviour for testing purposes
   """

   global test_mode
   global test_latest_int
   test_latest_int = seed
   test_mode = True


def switch_off_test_mode():
   """Subsequent calls to new_uid() will produce standard uuid values (default behaviour).

   note:
      this function will have no effect unless switch_on_test_mode() has previously been called
   """

   global test_mode
   test_mode = False


def new_uuid():
   """Returns a new uuid based on the time (to 100ns) & MAC address option of the iso standard.

   returns:
      uuid.UUID object

   notes:
      at present, the multi-processor safe functionality is not deployed, so multiple processors
      sharing the same MAC address could generate the same uuid simultaneously;
      an integer sequence is generated when in test mode
   """

   global test_latest_int
   if test_mode:
      test_latest_int += 1
      return uuid.UUID(bytes = test_latest_int.to_bytes(16, byte_order = 'big'))
   else:
      return uuid.uuid1()  # time to 100ns & MAC address


def string_from_uuid(uuid_obj):
   """Returns standard hexadecimal string for uuid; same as str(uuid_obj).

   arguments:
      uuid_obj (uuid.UUID object): the uuid which is required in hexadecimal string format

   returns:
      string (40 characters: 36 lowercase hexadecimals and 4 hyphens)
   """

   return str(uuid_obj)


def uuid_from_string(uuid_str):
   """Returns a uuid object for the given uuid string; hyphens are ignored.

   arguments:
      uuid_str (string): the hexadecimal representation of the 128 bit uuid integer;
         hyphens are ignored

   returns:
      uuid.UUID object

   notes:
      if a uuid.UUID object is passed by accident, it is returned;
      if the string starts with an underscore, the underscore is skipped (to cater for a fesapi quirk);
      any tail beyond the uuid string is ignored
   """

   if uuid_str is None:
      return None
   if isinstance(uuid_str, uuid.UUID):
      return uuid_str  # resilience to accidentally passing a uuid object
   try:
      if uuid_str[0] == '_':  # tolerate one of the fesapi quirks
         if len(uuid_str) < 37:
            return None
         return uuid.UUID(uuid_str[1:37])
      else:
         if len(uuid_str) < 36:
            return None
         return uuid.UUID(uuid_str[:36])
   except Exception:
      # could log or raise an error or warning?
      return None


def uuid_as_bytes(uuid_obj):
   """Returns the uuid as a 16 byte bytes sequence; same as uuid_obj.bytes.

   arguments:
      uuid_obj (uuid.UUID object): the uuid for which a bytes representation is required

   returns:
      bytes (16 bytes long)
   """

   if uuid_obj is None:
      return None
   if isinstance(uuid_obj, str):
      uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
   return uuid_obj.bytes


def uuid_as_int(uuid_obj):
   """Returns the uuid as a 128 bit int; same as uuid_obj.int.

   arguments:
      uuid_obj (uuid.UUID object): the uuid for which a bytes representation is required

   returns:
      bytes (16 bytes long)
   """

   if uuid_obj is None:
      return None
   if isinstance(uuid_obj, str):
      uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
   return uuid_obj.int


def matching_uuids(uuid_a, uuid_b):
   """Returns True if the 2 uuid objects are for the same id; False otherwise.

   arguments:
      uuid_a, uuid_b (uuid.UUID objects): the two uuids to be compared

   returns:
      boolean: True if the two uuids are the same; False otherwise

   note:
      this function is resilient to uuids being passed in hexadecimal string format
   """

   if isinstance(uuid_a, str):
      uuid_a = uuid_from_string(uuid_a)  # resilience to accidental string arg
   if isinstance(uuid_b, str):
      uuid_b = uuid_from_string(uuid_b)
   if uuid_a is None or uuid_b is None:
      return False
   return uuid_a.int == uuid_b.int


def version_string(uuid_obj):
   """Returns an integer string rendering of the time element of the uuid.

   arguments:
      uuid_obj (uuid.UUID): the uuid for which a string representation of the time component is required

   returns:
      string (of digits)

   notes:
      this function has nothing to do with the uuid.version attribute, it is used to populate the version
      field of a resqml citation block;
      the time component of the uuid is the number of 100ns periods that have elapsed since October 1582
      (when the Gregorian calendar was adopted), as a 60 bit integer
   """

   if isinstance(uuid_obj, str):
      uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
   v_str = str(uuid_obj.time)
   if len(v_str) > max_version_string_length:
      v_str = v_str[:max_version_string_length]
   return v_str



# resqpy test module including well blocking of trajectories against challenging grid geometries

version = '29th April 2021'

import pytest

import logging

log = logging.getLogger(__name__)

import os
import math as maths
import numpy as np
import pandas as pd

import resqpy.model as rq
import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.well as rqw
import resqpy.rq_import as rqi
import resqpy.olio.vector_utilities as vec
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet


def test_s_bend_fn(tmp_path, epc = None):

   if epc is None:
      # use pytest temporary directory fixture
      # https://docs.pytest.org/en/stable/tmpdir.html
      epc = str(os.path.join(tmp_path, f"{bu.new_uuid()}.epc"))

   # create s-bend grid

   nk = 5
   nj = 12
   ni_tail = 5
   ni_bend = 18
   ni_half_mid = 2
   ni = 2 * (ni_tail + ni_bend + ni_half_mid)

   total_thickness = 12.0
   layer_thickness = total_thickness / float(nk)
   flat_dx_di = 10.0
   horst_dx_dk = 0.25 * flat_dx_di / float(nk)
   horst_dz = 1.73 * layer_thickness
   horst_half_dx = horst_dx_dk * horst_dz / layer_thickness
   dy_dj = 8.0
   top_depth = 100.0

   assert ni_bend % 2 == 0, 'ni_bend must be even for horizontal faulting'
   assert ni_tail >= 5, 'ni_tail must be at least 5 for horst blocks'

   bend_theta_di = maths.pi / float(ni_bend)
   outer_radius = 2.0 * total_thickness

   bend_a_centre_xz = (flat_dx_di * float(ni_tail), top_depth + outer_radius)
   bend_b_centre_xz = (flat_dx_di * float(ni_tail - 2.0 * ni_half_mid),
                       top_depth + 3.0 * outer_radius - total_thickness)

   points = np.empty((nk + 1, nj + 1, ni + 1, 3))

   for k in range(nk + 1):
      if k == nk // 2 + 1:
         points[k] = points[k - 1]  # pinched out layer
      else:
         for i in range(ni + 1):
            if i < ni_tail + 1:
               x = flat_dx_di * float(i)
               z = top_depth + float(k) * layer_thickness  # will introduce a thick layer after pinchout
            elif i < ni_tail + ni_bend:
               theta = (i - ni_tail) * bend_theta_di
               radius = outer_radius - float(k) * layer_thickness
               x = bend_a_centre_xz[0] + radius * maths.sin(theta)
               z = bend_a_centre_xz[1] - radius * maths.cos(theta)
            elif i < ni_tail + ni_bend + 2 * ni_half_mid + 1:
               x = flat_dx_di * float(ni_tail - (i - (ni_tail + ni_bend)))
               z = top_depth + 2.0 * outer_radius - float(k) * layer_thickness
            elif i < ni_tail + 2 * ni_bend + 2 * ni_half_mid:
               theta = (i - (ni_tail + ni_bend + 2 * ni_half_mid)) * bend_theta_di
               radius = outer_radius - float(nk - k) * layer_thickness
               x = bend_b_centre_xz[0] - radius * maths.sin(theta)
               z = bend_b_centre_xz[1] - radius * maths.cos(theta)
            else:
               x = flat_dx_di * float((i - (ni - ni_tail)) + ni_tail - 2 * ni_half_mid)
               if i == ni - 1 or i == ni - 4:
                  x += horst_dx_dk * float(k)
               elif i == ni - 2 or i == ni - 3:
                  x -= horst_dx_dk * float(k)
               z = top_depth + 4.0 * outer_radius + float(k) * layer_thickness - 2.0 * total_thickness
            points[k, :, i] = (x, 0.0, z)

   for j in range(nj + 1):
      points[:, j, :, 1] = dy_dj * float(j)

   model = rq.Model(epc_file = epc, new_epc = True, create_basics = True, create_hdf5_ext = True)

   grid = grr.Grid(model)

   crs = rqc.Crs(model)
   crs_node = crs.create_xml()
   if model.crs_root is None:
      model.crs_root = crs_node

   grid.grid_representation = 'IjkGrid'
   grid.extent_kji = np.array((nk, nj, ni), dtype = 'int')
   grid.nk, grid.nj, grid.ni = nk, nj, ni
   grid.k_direction_is_down = True  # dominant layer direction, or paleo-direction
   grid.pillar_shape = 'straight'
   grid.has_split_coordinate_lines = False
   grid.k_gaps = None
   grid.crs_uuid = crs.uuid
   grid.crs_root = crs_node

   grid.points_cached = points

   grid.geometry_defined_for_all_pillars_cached = True
   grid.geometry_defined_for_all_cells_cached = True
   grid.grid_is_right_handed = crs.is_right_handed_xyz()

   grid.write_hdf5_from_caches()
   grid.create_xml()

   # create a well trajectory and md datum

   def df_trajectory(x, y, z):
      N = len(x)
      assert len(y) == N and len(z) == N
      df = pd.DataFrame(columns = ['MD', 'X', 'Y', 'Z'])
      md = np.zeros(N)
      for n in range(N - 1):
         md[n + 1] = md[n] + vec.naive_length((x[n + 1] - x[n], y[n + 1] - y[n], z[n + 1] - z[n]))
      df.MD = md
      df.X = x
      df.Y = y
      df.Z = z
      return df

   x = np.array(
      [0.0, flat_dx_di * float(ni_tail) + outer_radius, flat_dx_di * (float(ni_tail) - 0.5), 0.0, -outer_radius])
   y = np.array([0.0, dy_dj * 0.5, dy_dj * float(nj) / 2.0, dy_dj * (float(nj) - 0.5), dy_dj * float(nj)])
   z = np.array([
      0.0, top_depth - total_thickness, top_depth + 2.0 * outer_radius - total_thickness / 2.0,
      top_depth + 3.0 * outer_radius - total_thickness, top_depth + 4.0 * outer_radius
   ])

   df = df_trajectory(x, y, z)

   datum = rqw.MdDatum(model, crs_uuid = crs.uuid, location = (x[0], y[0], z[0]))
   datum.create_xml()

   trajectory = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'ANGLED_WELL')

   rqc.Crs(model, uuid = rqet.uuid_for_part_root(trajectory.crs_root)).uuid == crs.uuid

   trajectory.write_hdf5()
   trajectory.create_xml()

   # add more wells

   x = np.array(
      [0.0, flat_dx_di * float(ni_tail), flat_dx_di * 2.0 * float(ni_tail - ni_half_mid) + outer_radius, -outer_radius])
   y = np.array([0.0, dy_dj * float(nj) * 0.59, dy_dj * 0.67, dy_dj * 0.5])
   z = np.array([
      0.0, top_depth - total_thickness, top_depth + 4.0 * outer_radius - 1.7 * total_thickness,
      top_depth + 4.0 * outer_radius - 1.7 * total_thickness
   ])

   df = df_trajectory(x, y, z)

   traj_2 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'HORST_WELL')
   traj_2.write_hdf5()
   traj_2.create_xml()
   traj_2.control_points

   x = np.array([0.0, 0.0, 0.0])
   y = np.array([0.0, dy_dj * float(nj) * 0.53, dy_dj * float(nj) * 0.53])
   z = np.array([0.0, top_depth - total_thickness, top_depth + 4.0 * outer_radius])

   df = df_trajectory(x, y, z)

   traj_3 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'VERTICAL_WELL')
   traj_3.write_hdf5()
   traj_3.create_xml()
   traj_3.control_points

   n_x = flat_dx_di * float(ni_tail) * 0.48
   n_y = dy_dj * float(nj) / 9.1
   o_y = -dy_dj * 0.45
   nd_x = n_y / 3.0
   x = np.array([
      0.0, n_x, n_x, n_x + nd_x, n_x + 2.0 * nd_x, n_x + 3.0 * nd_x, n_x + 4.0 * nd_x, n_x + 5.0 * nd_x,
      n_x + 6.0 * nd_x, n_x + 7.0 * nd_x, n_x + 8.0 * nd_x, n_x + 8.0 * nd_x
   ])
   y = np.array([
      0.0, o_y, o_y + n_y, o_y + 2.0 * n_y, o_y + 3.0 * n_y, o_y + 4.0 * n_y, o_y + 5.0 * n_y, o_y + 6.0 * n_y,
      o_y + 7.0 * n_y, o_y + 8.0 * n_y, o_y + 9.0 * n_y, o_y + 10.0 * n_y
   ])
   n_z1 = top_depth + total_thickness * 0.82
   n_z2 = top_depth - total_thickness * 0.17
   z = np.array([0.0, n_z1, n_z1, n_z2, n_z2, n_z1, n_z1, n_z2, n_z2, n_z1, n_z1, n_z2])

   df = df_trajectory(x, y, z)

   traj_4 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'NESSIE_WELL')
   traj_4.write_hdf5()
   traj_4.create_xml()
   traj_4.control_points

   # block wells against grid geometry

   log.info('unfaulted grid blocking of well ' + str(rqw.well_name(trajectory)))
   bw = rqw.BlockedWell(model, grid = grid, trajectory = trajectory)
   bw.write_hdf5()
   bw.create_xml()
   assert bw.cell_count == 19

   log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_2)))
   bw_2 = rqw.BlockedWell(model, grid = grid, trajectory = traj_2)
   bw_2.write_hdf5()
   bw_2.create_xml()
   assert bw_2.cell_count == 33

   log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_3)))
   bw_3 = rqw.BlockedWell(model, grid = grid, trajectory = traj_3)
   bw_3.write_hdf5()
   bw_3.create_xml()
   assert bw_3.cell_count == 18

   log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_4)))
   bw_4 = rqw.BlockedWell(model, grid = grid, trajectory = traj_4)
   bw_4.write_hdf5()
   bw_4.create_xml()
   assert bw_4.cell_count == 26

   # derive a faulted version of the grid

   cp = grid.corner_points(cache_cp_array = True).copy()

   # IK plane faults
   cp[:, 3:, :, :, :, :, :] += (flat_dx_di * 0.7, 0.0, layer_thickness * 1.3)
   cp[:, 5:, :, :, :, :, :] += (flat_dx_di * 0.4, 0.0, layer_thickness * 0.9)
   cp[:, 8:, :, :, :, :, :] += (flat_dx_di * 0.3, 0.0, layer_thickness * 0.6)

   # JK plane faults
   cp[:, :, ni_tail + ni_bend // 2:, :, :, :, 0] -= flat_dx_di * 0.57  # horizontal break mid top bend
   cp[:, :, ni_tail + ni_bend + ni_half_mid:, :, :, :, 2] += layer_thickness * 1.27  # vertical break in mid section

   # zig-zag fault
   j_step = nj // (ni_tail - 2)
   for i in range(ni_tail - 1):
      j_start = i * j_step
      if j_start >= nj:
         break
      cp[:, j_start:, i, :, :, :, 2] += 1.1 * total_thickness

   # JK horst blocks
   cp[:, :, ni - 4, :, :, :, :] -= (horst_half_dx, 0.0, horst_dz)
   cp[:, :, ni - 3:, :, :, :, 0] -= 2.0 * horst_half_dx
   cp[:, :, ni - 2, :, :, :, :] += (-horst_half_dx, 0.0, horst_dz)
   cp[:, :, ni - 1:, :, :, :, 0] -= 2.0 * horst_half_dx

   # JK horst block mid lower bend
   bend_horst_dz = horst_dz * maths.tan(bend_theta_di)
   cp[:, :,
      ni - (ni_tail + ni_bend // 2 + 1):ni - (ni_tail + ni_bend // 2 - 1), :, :, :, :] -= (horst_dz, 0.0, bend_horst_dz)
   cp[:, :, ni - (ni_tail + ni_bend // 2 - 1):, :, :, :, 2] -= 2.0 * bend_horst_dz

   faulted_grid = rqi.grid_from_cp(model,
                                   cp,
                                   crs.uuid,
                                   max_z_void = 0.01,
                                   split_pillars = True,
                                   split_tolerance = 0.01,
                                   ijk_handedness = 'right' if grid.grid_is_right_handed else 'left',
                                   known_to_be_straight = True)

   faulted_grid.write_hdf5_from_caches()
   faulted_grid.create_xml()

   # block wells against faulted grid

   log.info('faulted grid blocking of well ' + str(rqw.well_name(trajectory)))
   fbw = rqw.BlockedWell(model, grid = faulted_grid, trajectory = trajectory)
   fbw.write_hdf5()
   fbw.create_xml()
   assert fbw.cell_count == 32

   log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_2)))
   fbw_2 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_2)
   fbw_2.write_hdf5()
   fbw_2.create_xml()
   assert fbw_2.cell_count == 26

   log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_3)))
   fbw_3 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_3)
   fbw_3.write_hdf5()
   fbw_3.create_xml()
   assert fbw_3.cell_count == 14

   log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_4)))
   fbw_4 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_4)
   fbw_4.write_hdf5()
   fbw_4.create_xml()
   assert fbw_4.cell_count == 16

   # create a version of the faulted grid with a k gap

   k_gap_grid = rqi.grid_from_cp(model,
                                 cp,
                                 crs.uuid,
                                 max_z_void = 0.01,
                                 split_pillars = True,
                                 split_tolerance = 0.01,
                                 ijk_handedness = 'right' if grid.grid_is_right_handed else 'left',
                                 known_to_be_straight = True)

   # convert second layer to a K gap
   k_gap_grid.nk_plus_k_gaps = k_gap_grid.nk
   k_gap_grid.nk -= 1
   k_gap_grid.extent_kji[0] = k_gap_grid.nk
   k_gap_grid.k_gaps = 1
   k_gap_grid.k_gap_after_array = np.zeros(k_gap_grid.nk - 1, dtype = bool)
   k_gap_grid.k_gap_after_array[0] = True
   k_gap_grid.k_raw_index_array = np.zeros(k_gap_grid.nk, dtype = int)
   for k in range(1, k_gap_grid.nk):
      k_gap_grid.k_raw_index_array[k] = k + 1

   # clear some attributes which may no longer be valid
   k_gap_grid.pinchout = None
   k_gap_grid.inactive = None
   k_gap_grid.grid_skin = None
   if hasattr(k_gap_grid, 'array_thickness'):
      delattr(k_gap_grid, 'array_thickness')

   k_gap_grid.write_hdf5_from_caches()
   k_gap_grid.create_xml()

   k_gap_grid_uuid = k_gap_grid.uuid

   # reload k gap grid object to ensure it is properly initialised

   k_gap_grid = None
   k_gap_grid = grr.Grid(model, uuid = k_gap_grid_uuid)

   # block wells against faulted grid with k gap

   log.info('k gap grid blocking of well ' + str(rqw.well_name(trajectory)))
   try:
      gbw = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = trajectory)
      gbw.write_hdf5()
      gbw.create_xml()
      assert gbw.cell_count == 24
   except Exception:
      log.exception('failed to block well against k gap grid')

   log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_2)))
   try:
      gbw_2 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_2)
      gbw_2.write_hdf5()
      gbw_2.create_xml()
      assert gbw_2.cell_count == 20
   except Exception:
      log.exception('failed to block well against k gap grid')

   log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_3)))
   try:
      gbw_3 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_3)
      gbw_3.write_hdf5()
      gbw_3.create_xml()
      assert gbw_3.cell_count == 10
   except Exception:
      log.exception('failed to block well against k gap grid')

   log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_4)))
   try:
      gbw_4 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_4)
      gbw_4.write_hdf5()
      gbw_4.create_xml()
      assert gbw_4.cell_count == 10
   except Exception:
      log.exception('failed to block well against k gap grid')

   # store model

   model.store_epc()

   assert k_gap_grid.k_gaps
   assert len(k_gap_grid.k_raw_index_array) == k_gap_grid.nk
   assert len(k_gap_grid.k_gap_after_array) == k_gap_grid.nk - 1
   assert k_gap_grid.pinched_out((1, 0, 2))

   # clean up
   model.h5_release()
   os.remove(model.h5_file_name())



"""xml_et.py: Resqml xml element tree utilities module."""

version = '20th July 2021'

import logging

log = logging.getLogger(__name__)
log.debug('xml_et.py version ' + version)

import os

# import xml element tree parse method and classes here to allow single point for switching between lxml and etree
# alternative to lxml.etree: xml.etree.ElementTree
from lxml.etree import parse, ElementTree, Element, SubElement, _Element  # noqa

import resqpy.olio.uuid as bu

from resqpy.olio.xml_namespaces import namespace as ns, inverse_namespace as inv_ns, curly_namespace as cns

pretend_to_be_fesapi = False
use_fesapi_quirks = True
use_tabs = False

if use_fesapi_quirks:
   null_xml_text = '\n'
else:
   null_xml_text = ''


def strip_path(full_path):
   """Returns the filename part of full_path with any directory path removed.

      :meta private:
   """

   return os.path.basename(full_path)


def stripped_of_prefix(prefixed):
   """Returns a simplified version of an xml tag or other element with any {xsd defining prefix} stripped off."""

   if prefixed is None:
      return None
   if '}' in prefixed:
      return prefixed[prefixed.rfind('}') + 1:]
   return prefixed[prefixed.rfind(':') + 1:]


def colon_prefixed(curly_prefixed):
   """Returns a version of an xml tag with {url} prefix replaced with nsi: equivalent; also returns the nsi prefix."""

   if not curly_prefixed:
      return None, None
   if curly_prefixed[0] != '{':
      colon = curly_prefixed.find(':')
      if colon < 0:
         return curly_prefixed, None
      return curly_prefixed, curly_prefixed[:colon]
   pre_end = curly_prefixed.rfind('}')
   try:
      pre_colon = inv_ns[curly_prefixed[1:pre_end]]
   except Exception:
      return curly_prefixed, None
   return pre_colon + ':' + curly_prefixed[pre_end + 1:], pre_colon


def find_tag(root, tag_name, must_exist = False):
   """Finds the first child in xml node with a (prefix-stripped) tag matching given tag name."""

   if root is None:
      return None
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         return child
   if must_exist:
      raise ValueError(f"Expected tag {tag_name} not found in root {root}")
   return None


def find_tag_text(root, tag_name, must_exist = False):
   """Finds the first child in xml node with a tag matching given tag name; returns stripped text field."""

   if root is None:
      return None
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         return node_text(child)
   if must_exist:
      raise ValueError(f"Expected tag {tag_name} not found in root {root}")
   return None


def find_tag_bool(root, tag_name, must_exist = False):
   """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as bool."""

   if root is None:
      return None
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         return node_bool(child)
   if must_exist:
      raise ValueError(f"Expected tag {tag_name} not found in root {root}")
   return None


def find_tag_int(root, tag_name, must_exist = False):
   """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as int."""

   if root is None:
      return None
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         return node_int(child)
   if must_exist:
      raise ValueError(f"Expected tag {tag_name} not found in root {root}")
   return None


def find_tag_float(root, tag_name, must_exist = False):
   """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as float."""

   if root is None:
      return None
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         return node_float(child)
   if must_exist:
      raise ValueError(f"Expected tag {tag_name} not found in root {root}")
   return None


def find_nested_tags(root, tag_list):
   """Follows a list of tags in a nested xml hierarchy, returning the node at the deepest level."""

   if not tag_list:
      return None
   head = find_tag(root, tag_list[0])
   if head is None:
      return None
   if len(tag_list) == 1:
      return head
   return find_nested_tags(head, tag_list[1:])


def find_nested_tags_cast(root, tag_list, dtype = None):
   """Return value of nested tags as desired dtype.
   
   Follows a list of tags in a nested xml hierarchy, returning the stripped text
   of the node at the deepest level.
   """

   cast_func = {
      int: node_int,
      float: node_float,
      bool: node_bool,
      str: node_text,
      None: lambda x: x,
   }[dtype]

   node = find_nested_tags(root, tag_list)
   return cast_func(node)


def find_nested_tags_text(root, tag_list):
   """Follows a list of tags in a nested xml hierarchy, returning the stripped text of the node at the deepest level."""

   node = find_nested_tags(root, tag_list)
   return node_text(node)


def find_nested_tags_bool(root, tag_list):
   """Follows a list of tags in a nested xml hierarchy, returning the text of the node at the deepest level, as bool."""

   node = find_nested_tags(root, tag_list)
   return node_bool(node)


def find_nested_tags_int(root, tag_list):
   """Follows a list of tags in a nested xml hierarchy, returning the text of the node at the deepest level, as int."""

   node = find_nested_tags(root, tag_list)
   return node_int(node)


def find_nested_tags_float(root, tag_list):
   """Follows a list of tags in a nested xml hierarchy, returning the text of the node at the deepest level, as float."""

   node = find_nested_tags(root, tag_list)
   return node_float(node)


def count_tag(root, tag_name):
   """Returns the number of children in xml node with a (prefix-stripped) tag matching given tag name."""

   if root is None:
      return None
   count = 0
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         count += 1
   return count


def list_of_tag(root, tag_name):
   """Returns a list of children in xml node with a (prefix-stripped) tag matching given tag name."""

   if root is None:
      return None
   results = []
   for child in root:
      if stripped_of_prefix(child.tag) == tag_name:
         results.append(child)
   return results


def list_of_descendant_tag(root, tag_name):
   """Returns a list of descendants in xml node tree with a (prefix-stripped) tag matching given tag name."""

   if root is None:
      return None
   results = []
   for child in root.iterdescendants():
      if stripped_of_prefix(child.tag) == tag_name:
         results.append(child)
   return results


def list_obj_references(root, skip_hdf5 = True):
   """Returns list of nodes of type DataObjectReference."""

   if root is None:
      return None
   results = []
   if node_type(root) == 'DataObjectReference' and not (skip_hdf5 and stripped_of_prefix(root.tag) == 'HdfProxy'):
      results.append(root)
   for child in root:
      results += list_obj_references(child, skip_hdf5 = skip_hdf5)
   return results


def cut_obj_references(root, uuids_to_be_cut):
   """Deletes any object reference nodes to uuids in given list."""

   if root is None or not uuids_to_be_cut:
      return
   for child in root:
      if node_type(child) == 'DataObjectReference':
         referred_uuid = bu.uuid_from_string(find_tag_text(child, 'UUID', must_exist = True))
         for cut_uuid in uuids_to_be_cut:
            if bu.matching_uuids(referred_uuid, cut_uuid):
               root.remove(child)
               break
      else:
         cut_obj_references(child, uuids_to_be_cut)


def cut_nodes_of_types(root, types_to_be_cut):
   """Deletes any nodes of a type matching one in the given list."""

   if root is None or not types_to_be_cut:
      return
   for child in root:
      if node_type(child) in types_to_be_cut:
         root.remove(child)  # hope this doesn't mess up the iteration
      else:
         cut_nodes_of_types(child, types_to_be_cut)


def content_type(content_type_str):
   """Returns the actual type, as embedded in an xml ContentType attribute; application and version are disregarded."""

   if content_type_str is None:
      return None
   if 'type=' in content_type_str:
      return content_type_str[content_type_str.rfind('type=') + 5:]


#   if ':' in content_type_str:
#      return content_type_str[content_type_str.rfind(':') + 1:]
   return content_type_str


def node_type(node, is_rels = False, strip_obj = False):
   """Returns the type as held in attributes of xml node; defining authority is stripped out."""

   result = None
   if node is None:
      return None
   if is_rels:
      if 'Type' not in node.attrib.keys():
         return None
      type_str = node.attrib['Type']
      result = type_str[type_str.rfind('/') + 1:]
   else:
      for key in node.attrib.keys():
         if stripped_of_prefix(key) == 'type':
            #           type_str = node.attrib[key]
            #           return type_str[type_str.rfind(':') + 1:]
            type_str = stripped_of_prefix(node.attrib[key])
            result = type_str
   if result and strip_obj and result.startswith('obj_'):
      result = result[4:]
   return result


def print_xml_tree(root,
                   level = 0,
                   max_level = None,
                   strip_tag_refs = True,
                   to_log = False,
                   log_level = None,
                   max_lines = 0,
                   line_count = 0):
   """Print an xml tree in an indented semi-readable format; return accumulated number of lines."""

   if root is None or (max_level is not None and level > max_level):
      return line_count
   if log_level is not None:
      to_log = True
   if log_level is None or log_level == 'info':
      print_fn = log.info
   elif log_level == 'debug':
      print_fn = log.debug
   elif log_level in ['warn', 'warning']:
      print_fn = log.warning
   elif log_level == 'error':
      print_fn = log.error
   else:
      print_fn = log.critical
   if max_lines and line_count >= max_lines:
      if line_count == max_lines:
         message = '(...xml tree output truncated after ' + str(max_lines) + ' lines)'
         if to_log:
            print_fn(message)
         else:
            print(message)
         line_count += 1
      return line_count

   value = root.text
   type_attr = None
   if strip_tag_refs:
      tag = stripped_of_prefix(root.tag)
      attrib_dict = {}
      for key in root.attrib.keys():
         stripped_key = stripped_of_prefix(key)
         attrib_dict[stripped_key] = root.attrib[key]
         if stripped_key == 'type':
            type_attr = stripped_of_prefix(root.attrib[key])
   else:
      tag = root.tag
      attrib_dict = root.attrib
      for key in root.attrib.keys():
         if stripped_of_prefix(key) == 'type':
            type_attr = root.attrib[key]
   if to_log:
      message = (3 * level * ' ') + tag + ' # ' + str(attrib_dict)
      if type_attr is not None:
         message += ' ## ' + type_attr
      if value is not None:
         message += ' ### ' + root.text.replace('\n', '\\n')
      print_fn(message)
   else:
      print((3 * level * ' ') + tag, '#', attrib_dict, end = ' ')
      if type_attr is not None:
         print('##', type_attr, end = ' ')
      if value is not None:
         print('###', root.text.replace('\n', '\\n'), end = '')
      print('')
   line_count += 1
   for child in root:
      line_count = print_xml_tree(child,
                                  level = level + 1,
                                  max_level = max_level,
                                  strip_tag_refs = strip_tag_refs,
                                  to_log = to_log,
                                  log_level = log_level,
                                  max_lines = max_lines,
                                  line_count = line_count)
      if line_count > max_lines:
         break
   return line_count


def uuid_in_part_name(part_name):
   """Returns uuid as embedded in part name."""

   # This might not always work
   if part_name is None:
      return None
   if part_name.endswith('.xml') and len(part_name) >= 40:
      return bu.uuid_from_string(part_name[-40:-4])
   elif part_name.endswith('.xml.rels') and len(part_name) >= 45:
      return bu.uuid_from_string(part_name[-45:-9])
   return None


def part_name_for_object(obj_type, uuid, prefixed = False, epc_subdir = None):
   """Returns the standard part name comprised of the object type, uuid and .xml extension."""

   if prefixed and (pretend_to_be_fesapi or use_fesapi_quirks) and obj_type[0] != '/':
      prefix = '/'
   else:
      prefix = ''
   if not obj_type.startswith('obj_'):
      prefix += 'obj_'
   if epc_subdir:
      if not epc_subdir.endswith('/'):
         epc_subdir += '/'
      prefix = epc_subdir + prefix
   return prefix + obj_type + '_' + str(uuid) + '.xml'


def rels_part_name_for_part(part_name):
   """Returns the paired relationships part name for the given part name."""

   pn = stripped_of_prefix(part_name)
   if pn is None or len(pn) == 0:
      return None
   dir_place = pn.rfind('/')
   if dir_place == -1:
      return '_rels/' + pn + '.rels'
   if dir_place == 0:
      return '_rels' + pn + '.rels'
   if dir_place == len(pn) - 1:
      return None
   return pn[:dir_place + 1] + '_rels' + pn[dir_place:] + '.rels'


def uuid_for_part_root(root):
   """Returns uuid as stored in xml attribs for root."""

   if root is None:
      return None
   if 'uuid' not in root.attrib.keys():
      return None
   return bu.uuid_from_string(root.attrib['uuid'])


def patch_uuid_in_part_root(root, uuid):
   """Returns modified part name with uuid swapped to uuid argument; root attrib is also changed."""

   if root is None or uuid is None:
      return None
   # This might not always work
   root.attrib['uuid'] = str(uuid)
   return part_name_for_part_root(root)


def part_name_for_part_root(root, is_rels = False, epc_subdir = None):
   """Returns the part name given the root node for the part's xml."""

   if root is None:
      return None
   obj_type = node_type(root, is_rels = is_rels)
   uuid = uuid_for_part_root(root)
   if obj_type is None or uuid is None:
      return None
   return part_name_for_object(obj_type, uuid, epc_subdir = epc_subdir)


# the next two functions aren't really much to do with the xml element tree


def find_in_ordered_data(value, array_1d):
   """Returns the index in the ordered list-like array of value; or None if not present."""

   def find_in_subset(value, array_1d, start, end):
      # recursive binary split
      if start >= end:
         return None
      mid = start + (end - start) // 2
      sample = array_1d[mid]
      if sample == value:
         while mid > 0 and array_1d[mid - 1] == value:
            mid -= 1
         return mid
      if sample > value:
         return find_in_subset(value, array_1d, start, mid)
      if mid == start:
         mid += 1
      return find_in_subset(value, array_1d, mid, end)

   return find_in_subset(value, array_1d, 0, len(array_1d))


def simplified_data_type(array_dtype):
   """Returns a simplified string version of the elemental data type (typically for a numpy or hdf5 array)."""

   str_dtype = str(array_dtype)
   if str_dtype.startswith('int'):
      return 'int'
   if str_dtype.startswith('float') or str_dtype.startswith('real'):
      return 'float'
   if str_dtype.startswith('bool'):
      return 'bool'
   return str_dtype


# following functions mostly previously in resqml_print.py


def bool_from_text(text):
   """Returns boolean value for string 'true' or 'false'; anything else results in None."""

   if text is None:
      return None
   if text.strip().lower() == 'true':
      return True
   if text.strip().lower() == 'false':
      return False
   return None


def node_text(node, unknown_if_none = False):
   """Returns stripped node text or 'unknown' if node is None or text is blank or newline."""

   if node is None or node.text is None:
      return 'unknown' if unknown_if_none else None
   text = node.text.strip()
   if len(text):
      return text
   return 'unknown' if unknown_if_none else None


def node_bool(node):
   """Returns stripped node text as bool, or None."""

   if node is None:
      return None
   return bool_from_text(node.text)


def node_int(node):
   """Returns stripped node text as int, or None."""

   if node is None:
      return None
   text = node.text.strip()
   if text.lower() == 'none':
      return None
   if len(text):
      return int(text)
   return None


def node_float(node):
   """Returns stripped node text as float, or None."""

   if node is None:
      return None
   text = node.text.strip()
   if text.lower() == 'none':
      return None
   if len(text):
      return float(text)
   return None


def length_units_from_node(node):
   """Returns standard length units string based on node text, or 'unknown'."""

   if node is None or node.text == '' or node.text == '\n':
      return 'unknown'
   else:
      return node.text.strip()


def time_units_from_node(node):
   """Returns standard time units string based on node text, or 'unknown'."""

   if node is None or node.text == '' or node.text == '\n':
      return 'unknown'
   else:
      return node.text.strip()


def xyz_handedness(xy_axes, z_inc_down):
   """Returns xyz true handedness as 'left', 'right' or 'unknown', based on xy_axes (string) and z_inc_down (boolean)."""

   if xy_axes is None or z_inc_down is None:
      return 'unknown'
   xy_axes_split = xy_axes.lower().split()
   if len(xy_axes_split) != 2:
      return 'unknown'
   if xy_axes_split not in [
      ['easting', 'northing'],
      ['northing', 'easting'],  # only these 6 options allowed in resqml
      ['westing', 'southing'],
      ['southing', 'westing'],
      ['northing', 'westing'],
      ['westing', 'northing']
   ]:
      return 'unknown'
   right_handed = z_inc_down
   if xy_axes_split in [['easting', 'northing'], ['westing', 'southing'], ['northing', 'westing']]:
      right_handed = not right_handed
   if right_handed:
      return 'right'
   else:
      return 'left'


def ijk_handedness(geom_node):
   """Returns ijk true handedness as 'left', 'right' or 'unknown', based on GridIsRightHanded node in grid geometry node."""

   if geom_node is None:
      return 'unknown'
   right_handed = bool_from_text(node_text(find_tag(geom_node, 'GridIsRighthanded')))
   if right_handed is None:
      return 'unknown'
   if right_handed:
      return 'right'
   return 'left'


def citation_title_for_node(node):
   """Looks for a citation node as a child of node and returns the title text."""

   return find_nested_tags_text(node, ['Citation', 'Title'])


def creation_date_for_node(node):
   """Looks for a citation node as a child of node and returns the creation (date-time) text."""

   return find_nested_tags_text(node, ['Citation', 'Creation'])


def write_xml_node(xml_fp, root, level = 0, namespace_keys = []):
   """Recursively write an xml node to an open file; return number of nodes written."""

   if root is None:
      return 0

   ns_keys = namespace_keys.copy()

   type_attr = None
   tag, pre_colon = colon_prefixed(root.tag)

   if pre_colon in ['content_types', 'rels']:
      tag = tag[len(pre_colon) + 1:]
      ct_special = True
   else:
      ct_special = False

   if use_tabs:
      line = (level * '\t')
   else:
      line = (3 * level * ' ')
   line += '<' + tag
   if pre_colon and pre_colon not in ns_keys:
      line += ' xmlns'
      if not ct_special:
         line += ':' + pre_colon
      line += '="' + ns[pre_colon] + '"'
      ns_keys.append(pre_colon)

   attrib_ns_list = []
   attrib_list = []
   type_pre_colon = None
   for key in root.attrib.keys():
      colon_attrib_key, pre_colon_attrib = colon_prefixed(key)
      if pre_colon_attrib and pre_colon_attrib not in ns_keys:
         attrib_ns_list.append(pre_colon_attrib)
      if stripped_of_prefix(key) == 'type':
         type_attr, type_pre_colon = colon_prefixed(root.attrib[key])
         attrib_list.append(colon_attrib_key + '="' + type_attr + '"')
      elif ct_special and colon_attrib_key == 'PartName' and root.attrib[key].startswith('obj_'):
         attrib_list.append(colon_attrib_key + '="/' + root.attrib[key] + '"')
      else:
         attrib_list.append(colon_attrib_key + '="' + root.attrib[key] + '"')

   for attrib_ns in attrib_ns_list:
      line += ' xmlns:' + attrib_ns + '="' + ns[attrib_ns] + '"'
      ns_keys.append(attrib_ns)
   if type_pre_colon and type_pre_colon not in ns_keys:  # must be included in the local xml line?
      line += ' xmlns:' + type_pre_colon + '="' + ns[type_pre_colon] + '"'
      ns_keys.append(type_pre_colon)

   for attrib in attrib_list:
      line += ' ' + attrib

   node_count = 1

   if ct_special and len(root) == 0:

      line += '/>\n'
      xml_fp.write(line.encode())
#      print(line, end = '') # debug

   else:

      line += '>'

      if root.text and not root.text.isspace():
         line += root.text

      xml_fp.write(line.encode())
      #      print(line, end = '') # debug
      indentation = ''

      if len(root):
         xml_fp.write(b'\n')
         #         print()  # debug
         if use_tabs:
            indentation = (level * '\t')
         else:
            indentation = (3 * level * ' ')
         for child in root:
            node_count += write_xml_node(xml_fp, child, level = level + 1, namespace_keys = ns_keys)

      line = indentation + '</' + tag + '>\n'
      xml_fp.write(line.encode())


#      print(line, end = '') # debug

   return node_count


def write_xml(xml_fp, tree, standalone = None):
   """Write an xml tree to file in an indented format; gSOAP/FESAPI compatible; return number of nodes written."""

   #   print('-------------------------------------------------------------------------------------') # debug
   line = '<?xml version="1.0" encoding="UTF-8"'
   if standalone is not None:
      line += ' standalone="' + standalone + '"'
   line += '?>\n'
   xml_fp.write(line.encode())
   #   print(line, end = '')  # debug
   nodes = write_xml_node(xml_fp, tree.getroot())

   return nodes


def load_metadata_from_xml(node):
   """Loads the ExtraMetaData stored in a RESQML part as a dictionary"""

   if node is None:
      return None
   extra_metadata = {}
   meta_nodes = list_of_tag(node, 'ExtraMetadata')
   for meta in meta_nodes:
      name = find_tag_text(meta, 'Name')
      value = find_tag_text(meta, 'Value')
      extra_metadata[name] = value
   return extra_metadata


def create_metadata_xml(node, extra_metadata):
   """Writes the xml for the given metadata dictionary"""

   if extra_metadata:
      for data in extra_metadata.keys():

         metadata = SubElement(node, cns['resqml2'] + 'ExtraMetadata')
         metadata.set(cns['xsi'] + 'type', cns['resqml2'] + 'NameValuePair')
         metadata.text = null_xml_text

         name = SubElement(metadata, cns['resqml2'] + 'Name')
         name.set(cns['xsi'] + 'type', cns['xsd'] + 'string')
         name.text = str(data)

         value = SubElement(metadata, cns['resqml2'] + 'Value')
         value.set(cns['xsi'] + 'type', cns['xsd'] + 'string')
         value.text = str(extra_metadata[data])

   return node


def is_node(obj):
   """Returns True if type of object is element tree node; False otherwise."""

   # note: only tested for lxml

   return type(obj) is _Element or type(obj) is Element



# keyword_files module
# provides basic functions for searching for keywords in an ascii control file such as a nexus deck
# ascii file must already have been opened for reading before calling any of these functions

version = '4th September 2020'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)
log.debug('keyword_files.py version %s', version)

# defs:
# def substring(shorter, longer):
# def end_of_file(ascii_file):
# def find_keyword(ascii_file, keyword, max_lines = None):
# def skip_blank_lines_and_comments(ascii_file, comment_intro = '!'):
# def split_trailing_comment(line, comment_char = '!'):
# def strip_trailing_comment(line, comment_char = '!'):
# def find_keyword_without_passing(ascii_file, keyword, no_pass_keyword):
# def find_keyword_with_copy(ascii_file_in, keyword, ascii_file_out):
# def find_keyword_pair(ascii_file, primary_keyword, secondary_keyword):
# def find_number(ascii_file):
# def specific_keyword_next(ascii_file, keyword, skip_blank_lines = True, comment_char = '!'):
# def number_next(ascii_file, skip_blank_lines = True, comment_char = '!'):
# def guess_comment_char(ascii_file):


def substring(shorter, longer):
   """Returns True if the first argument is a substring of the second."""
   try:
      longer.index(shorter)
      return True
   except Exception:
      return False


def end_of_file(ascii_file):
   """Returns True if the end of the file has been reached."""
   file_pos = ascii_file.tell()
   line = ascii_file.readline()
   if len(line) == 0:
      return True  # end of file
   ascii_file.seek(file_pos)
   return False


def find_keyword(ascii_file, keyword, max_lines = None):
   """Looks for line starting with given keyword; file pointer is left at start of that line."""
   start_pos = ascii_file.tell()
   while True:
      if max_lines is not None:
         if max_lines <= 0:
            ascii_file.seek(start_pos)
            return False
         max_lines -= 1
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         ascii_file.seek(start_pos)
         return False  # end of file
      words = line.split()
      if len(words) > 0 and words[0].upper() == keyword.upper():
         ascii_file.seek(file_pos)
         return True


def skip_blank_lines_and_comments(ascii_file, comment_char = '!', skip_c_space = True):
   """Skips any lines containing only white space or comment."""
   while True:
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         return  # end of file
      words = line.split()
      if len(words) == 0:
         continue
      if words[0][0] == comment_char:
         continue
      if skip_c_space and len(words[0]) == 1 and (words[0][0] == 'C' or words[0][0] == 'c'):
         continue
      ascii_file.seek(file_pos)
      return


def skip_comments(ascii_file, comment_char = '!', skip_c_space = True):
   """Skips any lines containing only a comment."""
   while True:
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         return  # end of file
      words = line.split()
      if len(words) > 0:  # not a blank line
         if words[0][0] == comment_char:
            continue
         if skip_c_space and len(words[0]) == 1 and (words[0][0] == 'C' or words[0][0] == 'c'):
            continue
      ascii_file.seek(file_pos)
      return


def split_trailing_comment(line, comment_char = '!'):
   """Returns a pair of strings: (line stripped of trailing comment, trailing comment)."""
   # also removes trailing newline
   comment = ''
   local_line = line
   while len(local_line) > 0 and local_line[-1] in ['\n', '\r']:
      local_line = local_line[:-1]
   try:
      pling = local_line.index(comment_char)
      if pling < len(local_line) - 2:
         comment = local_line[pling + 1:]
      return (local_line[:pling], comment)
   except Exception:
      return (local_line, '')


def strip_trailing_comment(line, comment_char = '!'):
   """Returns a copy of line with any trailing comment removed."""
   (result, comment) = split_trailing_comment(line, comment_char = comment_char)
   return result


def find_keyword_without_passing(ascii_file, keyword, no_pass_keyword):
   """Looks for line starting with keyword, but without passing line starting with no_pass_keyword."""
   while True:
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         return False  # end of file
      words = line.split()
      if len(words) > 0:
         if words[0].upper() == keyword.upper():
            ascii_file.seek(file_pos)
            return True
         if words[0].upper() == no_pass_keyword.upper():
            ascii_file.seek(file_pos)
            return False


def find_keyword_with_copy(ascii_file_in, keyword, ascii_file_out):
   """Looks for line starting with given keyword, copying lines in the meantime."""
   while True:
      file_pos = ascii_file_in.tell()
      line = ascii_file_in.readline()
      if len(line) == 0:
         return False  # end of file
      words = line.split()
      if len(words) > 0 and words[0].upper() == keyword.upper():
         ascii_file_in.seek(file_pos)
         return True
      else:
         ascii_file_out.write(line)


def find_keyword_pair(ascii_file, primary_keyword, secondary_keyword):
   """Looks for line starting with a given pair of keywords."""
   while True:
      if not find_keyword(ascii_file, primary_keyword):
         return False
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         return False  # end of file
      words = line.split()
      if len(words) >= 2 and words[1].upper() == secondary_keyword.upper():
         ascii_file.seek(file_pos)
         return True


def find_number(ascii_file):
   """Looks for line starting with any number."""
   while True:
      file_pos = ascii_file.tell()
      line = ascii_file.readline()
      if len(line) == 0:
         return False  # end of file
      words = line.split()
      if len(words) > 0:
         try:
            float(words[0])
         except Exception:  # todo: should only catch a particular exception (type conversion)
            continue
         else:
            ascii_file.seek(file_pos)
            return True


def specific_keyword_next(ascii_file, keyword, skip_blank_lines = True, comment_char = '!'):
   """Returns True if next token in file is the specified keyword."""
   file_pos = ascii_file.tell()  # will restore file pos to original, which may be before blank lines
   while True:
      if (skip_blank_lines):
         file_pos = ascii_file.tell()  # will advance file pos to next non-blank line
      line = ascii_file.readline()
      if len(line) == 0:  # end of file
         ascii_file.seek(file_pos)
         return False
      words = line.split()
      if len(words) == 0 or words[0][0] == comment_char:
         continue  # blank line or comment
      ascii_file.seek(file_pos)  # restore file position whether or not keyword match
      if words[0].upper() == keyword.upper():
         return True
      return False


def number_next(ascii_file, skip_blank_lines = True, comment_char = '!'):
   """Returns True if next token in file is a number."""
   file_pos = ascii_file.tell()
   while True:
      if (skip_blank_lines):
         file_pos = ascii_file.tell()  # will advance file pos to next non-blank line
      line = ascii_file.readline()
      if len(line) == 0:  # end of file
         ascii_file.seek(file_pos)
         return False
      words = line.split()
      if len(words) == 0 or words[0][0] == comment_char:
         continue  # blank line or comment
      ascii_file.seek(file_pos)  # restore file position whether or not number found
      try:
         float(words[0])
      except Exception:  # todo: should only catch a particular exception (type conversion)
         return False
      else:
         return True


def blank_line(ascii_file):
   """Returns True if the next line contains only white space; False otherwise (including comments)."""

   file_pos = ascii_file.tell()
   line = ascii_file.readline()
   ascii_file.seek(file_pos)
   if len(line) == 0:
      return True  # end of file
   words = line.split()
   return len(words) == 0


def guess_comment_char(ascii_file):
   """Returns a string (usually one character) being the guess as to the comment character, or None."""
   file_pos = ascii_file.tell()  # will restore file pos to original
   max_lines = 10
   ch = None
   while max_lines:
      line = ascii_file.readline()
      max_lines -= 1
      if not line:
         break
      words = line.split()
      if len(words) and words[0] in ['!', 'C', '#', '--']:
         ch = words[0]
         break
   ascii_file.seek(file_pos)
   return ch


# end of keyword_files module



from numpy.lib.arraysetops import isin
import pytest
import os
import numpy as np

import resqpy.model as rq
import resqpy.crs as rqc
import resqpy.well as rqw
import resqpy.property as rqp
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5


def test_model(tmp_path):

   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   assert model is not None
   crs = rqc.Crs(model)
   crs_root = crs.create_xml()
   model.store_epc()
   assert os.path.exists(epc)
   md_datum_1 = rqw.MdDatum(model, location = (0.0, 0.0, -50.0), crs_uuid = crs.uuid)
   md_datum_1.create_xml(title = 'Datum 1')
   md_datum_2 = rqw.MdDatum(model, location = (3.0, 0.0, -50.0), crs_uuid = crs.uuid)
   md_datum_2.create_xml(title = 'Datum 2')
   assert len(model.uuids(obj_type = 'MdDatum')) == 2
   model.store_epc()

   model = rq.Model(epc)
   assert model is not None
   assert len(model.uuids(obj_type = 'MdDatum')) == 2
   datum_part_1 = model.part(obj_type = 'MdDatum', title = '1', title_mode = 'ends')
   datum_part_2 = model.part(obj_type = 'MdDatum', title = '2', title_mode = 'ends')
   assert datum_part_1 is not None and datum_part_2 is not None and datum_part_1 != datum_part_2
   datum_uuid_1 = rqet.uuid_in_part_name(datum_part_1)
   datum_uuid_2 = rqet.uuid_in_part_name(datum_part_2)
   assert not bu.matching_uuids(datum_uuid_1, datum_uuid_2)
   p1 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_1)]
   p2 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_2)]
   assert p1 == datum_part_1 and p2 == datum_part_2


def test_model_iterators(example_model_with_well):

   model, well_interp, datum, traj = example_model_with_well

   w = next(model.iter_wellbore_interpretations())
   d = next(model.iter_md_datums())
   t = next(model.iter_trajectories())

   assert bu.matching_uuids(w.uuid, well_interp.uuid)
   assert bu.matching_uuids(d.uuid, datum.uuid)
   assert bu.matching_uuids(t.uuid, traj.uuid)
   assert w == well_interp
   assert d == datum
   assert t == traj


def test_model_iterate_objects(example_model_with_well):

   from resqpy.organize import WellboreFeature, WellboreInterpretation

   model, well_interp, _, _ = example_model_with_well

   # Try iterating over wellbore features

   wells = model.iter_objs(WellboreFeature)
   wells = list(wells)
   w = wells[0]
   assert len(wells) == 1
   assert isinstance(wells[0], WellboreFeature)
   assert wells[0].title == "well A"

   # Try iterating over wellbore interpretations

   interps = model.iter_objs(WellboreInterpretation)
   interps = list(interps)
   assert len(interps) == 1
   assert isinstance(interps[0], WellboreInterpretation)
   assert interps[0] == well_interp


def test_model_iter_crs(example_model_and_crs):

   model, crs_1 = example_model_and_crs

   crs_list = list(model.iter_crs())
   assert len(crs_list) == 1

   crs_2 = crs_list[0]
   assert crs_2 == crs_1


def test_model_iter_crs_empty(tmp_model):

   # Should raise an exception if no CRS exists
   with pytest.raises(StopIteration):
      next(tmp_model.iter_crs())


def test_model_as_graph(example_model_with_well):

   model, well_interp, datum, traj = example_model_with_well
   crs = next(model.iter_crs())

   nodes, edges = model.as_graph()

   # Check nodes
   for obj in [well_interp, datum, traj, crs]:
      assert str(obj.uuid) in nodes.keys()
      if hasattr(obj, 'title'):  # TODO: remove this when all objects have this attribute
         assert obj.title == nodes[str(obj.uuid)]['title']

   # Check edges
   assert frozenset([str(traj.uuid), str(datum.uuid)]) in edges
   assert frozenset([str(traj.uuid), str(crs.uuid)]) in edges
   assert frozenset([str(traj.uuid), str(well_interp.uuid)]) in edges
   assert frozenset([str(datum.uuid), str(traj.uuid)]) in edges

   # Test uuid subset
   nodes, edges = model.as_graph(uuids_subset = [datum.uuid])
   assert len(nodes.keys()) == 1
   assert len(edges) == 0


def test_model_copy_all_parts(example_model_with_properties):

   epc = example_model_with_properties.epc_file
   dir = example_model_with_properties.epc_directory
   copied_epc = os.path.join(dir, 'copied.epc')

   # test copying without consolidation
   original = rq.Model(epc)
   assert original is not None
   copied = rq.new_model(copied_epc)
   copied.copy_all_parts_from_other_model(original, consolidate = False)

   assert set(original.uuids()) == set(copied.uuids())
   assert set(original.parts()) == set(copied.parts())

   # test without consolidation of two crs objects
   copied = rq.new_model(copied_epc)
   new_crs = rqc.Crs(copied)
   new_crs.create_xml()

   copied.copy_all_parts_from_other_model(original, consolidate = False)

   assert len(copied.parts()) == len(original.parts()) + 1
   assert set(original.parts()).issubset(set(copied.parts()))
   assert len(copied.parts(obj_type = 'LocalDepth3dCrs')) == 2

   # test with consolidation of two crs objects
   copied = rq.new_model(copied_epc)
   new_crs = rqc.Crs(copied)
   new_crs.create_xml()

   copied.copy_all_parts_from_other_model(original, consolidate = True)

   assert len(copied.parts()) == len(original.parts())
   assert len(copied.parts(obj_type = 'LocalDepth3dCrs')) == 1

   crs_uuid = copied.uuid(obj_type = 'LocalDepth3dCrs')
   assert (bu.matching_uuids(crs_uuid, new_crs.uuid) or
           bu.matching_uuids(crs_uuid, original.uuid(obj_type = 'LocalDepth3dCrs')))

   # test write and re-load of copied model
   copied.store_epc()
   re_opened = rq.Model(copied_epc)
   assert re_opened is not None

   assert len(copied.parts()) == len(original.parts())

   crs_uuid = re_opened.uuid(obj_type = 'LocalDepth3dCrs')
   assert (bu.matching_uuids(crs_uuid, new_crs.uuid) or
           bu.matching_uuids(crs_uuid, original.uuid(obj_type = 'LocalDepth3dCrs')))


def test_model_copy_all_parts_non_resqpy_hdf5_paths(example_model_with_properties):
   # this test uses low level methods to override the hdf5 internal paths for some new objects

   epc = example_model_with_properties.epc_file
   dir = example_model_with_properties.epc_directory
   copied_epc = os.path.join(dir, 'copied.epc')
   extent_kji = (3, 5, 5)  # needs to match extent of grid in example model

   # add some properties with bespoke internal hdf5 paths
   original = rq.Model(epc)
   assert original is not None
   grid_uuid = original.uuid(obj_type = 'IjkGridRepresentation')
   assert grid_uuid is not None
   data = np.linspace(0.0, 1000.0, num = 75).reshape(extent_kji)

   hdf5_paths = [
      '/RESQML/uuid_waffle/values', '/RESQML/waffle_uuid/unusual_name', 'RESQML/class_uuid_waffle/values0',
      'RESQML/something_interesting/array', '/RESQML/abrupt', 'no_resqml/uuid/values'
   ]

   prop_list = []
   path_list = []  # elements will have 'uuid' replaced with uuid string
   h5_reg = rwh5.H5Register(original)
   for i, hdf5_path in enumerate(hdf5_paths):
      prop = rqp.Property(original, support_uuid = grid_uuid)
      if 'uuid' in hdf5_path:
         path = hdf5_path.replace('uuid', str(prop.uuid))
      else:
         path = hdf5_path
      prop.prepare_import(cached_array = data,
                          source_info = 'test data',
                          keyword = f'TEST{i}',
                          property_kind = 'continuous',
                          uom = 'm')
      for entry in prop.collection.imported_list:  # only one entry
         # override internal hdf5 path
         h5_reg.register_dataset(entry[0], 'values_patch0', prop.collection.__dict__[entry[3]], hdf5_path = path)
      prop_list.append(prop)
      path_list.append(path)
   h5_reg.write(mode = 'a')

   for prop, hdf5_path in zip(prop_list, path_list):
      root_node = prop.create_xml(find_local_property_kind = False)
      assert root_node is not None
      # override the hdf5 internal path in the xml tree
      path_node = rqet.find_nested_tags(root_node, ['PatchOfValues', 'Values', 'Values', 'PathInHdfFile'])
      assert path_node is not None
      path_node.text = hdf5_path

   # rewrite model
   original.store_epc()

   # re-open model and test copying
   original = rq.Model(epc)
   assert original is not None
   copied = rq.new_model(copied_epc)
   copied.copy_all_parts_from_other_model(original, consolidate = False)

   assert set(original.uuids()) == set(copied.uuids())
   assert set(original.parts()) == set(copied.parts())



# tests for resqpy.olio.fine_coarse class and functions

import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

import resqpy.olio.fine_coarse as rqfc


def test_fine_coarse():

   fc1 = rqfc.FineCoarse((3, 4, 5), (3, 4, 5))
   assert fc1 is not None
   fc1.set_all_ratios_constant()
   fc1.assert_valid()
   assert fc1.ratios((2, 2, 2)) == (1, 1, 1)

   fc2 = rqfc.FineCoarse((9, 16, 10), (3, 4, 5))
   assert fc2 is not None
   fc2.set_all_ratios_constant()
   fc2.set_all_proprtions_equal()
   fc2.assert_valid()
   assert fc2.ratios((2, 2, 2)) == (3, 4, 2)

   assert fc2.coarse_for_fine_kji0((5, 5, 5)) == (1, 1, 2)
   assert fc2.coarse_for_fine_axial(1, 13) == 3
   assert fc2.fine_base_for_coarse_axial(0, 1) == 3
   assert fc2.fine_base_for_coarse((2, 2, 2)) == (6, 8, 4)
   assert np.all(fc2.fine_box_for_coarse((1, 2, 3)) == np.array([[3, 8, 6], [5, 11, 7]], dtype = int))

   assert_array_almost_equal(fc2.proportion(1, 1), (0.25, 0.25, 0.25, 0.25))
   assert_array_almost_equal(fc2.interpolation(1, 1), (0.0, 0.25, 0.5, 0.75))

   # todo: several more methods to test


def test_fine_coarse_functions():

   assert rqfc.axis_for_letter('k') == 0
   assert rqfc.axis_for_letter('J') == 1
   assert rqfc.axis_for_letter('i') == 2
   assert rqfc.letter_for_axis(0) == 'K'
   assert rqfc.letter_for_axis(2) == 'I'

   fct = rqfc.tartan_refinement(coarse_extent_kji = (5, 7, 7),
                                coarse_fovea_box = np.array([[0, 2, 2], [1, 4, 4]], dtype = int),
                                fovea_ratios_kji = (3, 3, 4),
                                decay_rates_kji = (1, 1, 1),
                                decay_mode = 'linear')
   assert fct is not None
   fct.assert_valid()

   assert fct.coarse_extent_kji == (5, 7, 7)
   assert fct.fine_extent_kji == (10, 15, 22)
   assert np.all(fct.coarse_for_fine_axial_vector(0) == (0, 0, 0, 1, 1, 1, 2, 2, 3, 4))
   for const in fct.constant_ratios:
      assert const is None
   assert np.all(fct.vector_ratios[1] == (1, 2, 3, 3, 3, 2, 1))
   assert np.all(fct.vector_ratios[2] == (2, 3, 4, 4, 4, 3, 2))

   # todo: add test of tartan refinement in exponential mode



# module defining dictionary of nexus WELLSPEC column keywords

version = '29th April 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
from typing import Dict, Tuple, Any, Type

log = logging.getLogger(__name__)
log.debug('wellspec_keywords.py version %s', version)

import numpy as np
import pandas as pd

import resqpy.olio.keyword_files as kf

# functions defined here:
# def increment_complaints(keyword):
# def known_keyword(keyword):
# def add_unknown_keyword(keyword):
# def default_value(keyword):
# def complaints(keyword):
# def required_out_list():
# def length_unit_conversion_applicable(keyword):

# nexus wellspec columns as required by pagoda
wk_unknown = -3
wk_banned = -2
wk_preferably_not = -1
wk_okay = 0
wk_preferred = 1
wk_required = 2

wellspec_dict: Dict[str, Tuple[int, int, int, Any, bool]] = {}  # mapping wellspec column key to:
#     (warn count, required in, required out, default, length units boolean, )

# NB: changing entries in this list will usually require other code change elsewhere
# second element of tuple should be >= first element
# yapf: disable
wellspec_dict['IW']       = (0, wk_required,  wk_required,  None,   False)
wellspec_dict['JW']       = (0, wk_required,  wk_required,  None,   False)
wellspec_dict['L']        = (0, wk_required,  wk_required,  None,   False)
wellspec_dict['GRID']     = (0, wk_preferred, wk_required,  None,   False)  # or use main grid name as default
wellspec_dict['RADW']     = (0, wk_preferred, wk_required,  0.25,   True)   # use pagoda spec value for i.p. perf
wellspec_dict['KHMULT']   = (0, wk_okay,      wk_okay,      1.0,    False)  # or use 'NA" as default?
wellspec_dict['STAT']     = (0, wk_okay,      wk_okay,      'ON',   False)
wellspec_dict['ANGLA']    = (0, wk_preferred, wk_required,  0.0,    False)
wellspec_dict['ANGLV']    = (0, wk_preferred, wk_required,  0.0,    False)  # default for other perfs (vertical)
wellspec_dict['LENGTH']   = (0, wk_okay,      wk_okay,      None,   True)   # derive default from cell size
wellspec_dict['KH']       = (0, wk_okay,      wk_okay,      None,   True)   # althernative to LENGTH (one required)
wellspec_dict['SKIN']     = (0, wk_okay,      wk_okay,      0.0,    False)
wellspec_dict['PPERF']    = (0, wk_okay,      wk_okay,      1.0,    False)
wellspec_dict['ANGLE']    = (0, wk_okay,      wk_okay,      360.0,  False)
wellspec_dict['IRELPM']   = (0, wk_okay,      wk_okay,      None,   False)  # default fracture IRELPM for i.p. perf
wellspec_dict['RADB']     = (0, wk_okay,      wk_okay,      None,   True)   # calculate from cell size & k_align, k_v
wellspec_dict['WI']       = (0, wk_okay,      wk_okay,      None,   False)  # caluclate from radb, radw & skin
wellspec_dict['K']        = (0, wk_preferably_not, wk_okay, None,   False)  # derive from conductivity for i.p. perf?
wellspec_dict['LAYER']    = (0, wk_preferably_not, wk_okay, None,   False)  # use LGR i.p. layer for i.p. perf
wellspec_dict['DEPTH']    = (0, wk_okay,      wk_okay,      '#',    True)   # # causes nexus to use cell depth
wellspec_dict['X']        = (0, wk_okay,      wk_okay,      None,   True)   # use cell X for i.p. perf
wellspec_dict['Y']        = (0, wk_okay,      wk_okay,      None,   True)   # use cell Y for i.p. perf
wellspec_dict['CELL']     = (0, wk_banned,    wk_banned,    None,   False)  # CELL is for unstructured grids
wellspec_dict['DTOP']     = (0, wk_banned,    wk_banned,    None,   True)   # not compatible with ANGLA, ANGLV
wellspec_dict['DBOT']     = (0, wk_banned,    wk_banned,    None,   True)   # not compatible with ANGLA, ANGLV
wellspec_dict['RADBP']    = (0, wk_preferably_not, wk_okay, None,   True)   # calculate as for RADB
wellspec_dict['RADWP']    = (0, wk_preferably_not, wk_okay, None,   True)   # use pagoda wellbore radius
wellspec_dict['PORTYPE']  = (0, wk_banned,    wk_banned,    None,   False)  # dual porosity: todo: need to check values
wellspec_dict['FM']       = (0, wk_preferably_not, wk_okay, 0.0,    False)  # dual porosity: connection to fracture?
wellspec_dict['MD']       = (0, wk_preferably_not, wk_okay, 'NA',   False)
wellspec_dict['PARENT']   = (0, wk_preferably_not, wk_okay, 'NA',   False)
wellspec_dict['MDCON']    = (0, wk_preferably_not, wk_okay, 'NA',   False)
wellspec_dict['SECT']     = (0, wk_preferably_not, wk_okay, 1,      False)
wellspec_dict['FLOWSECT'] = (0, wk_preferably_not, wk_okay, 1,      False)
wellspec_dict['ZONE']     = (0, wk_preferably_not, wk_okay, 1,      False)
wellspec_dict['GROUP']    = (0, wk_preferably_not, wk_okay, 1,      False)
wellspec_dict['TEMP']     = (0, wk_preferably_not, wk_okay, 'NA',   False)
wellspec_dict['IPTN']     = (0, wk_preferably_not, wk_okay, 1,      False)  # pattern
wellspec_dict['D']        = (0, wk_preferably_not, wk_okay, 'NA',   False)  # non D'Arcy flow
wellspec_dict['ND']       = (0, wk_preferably_not, wk_okay, 'NA',   False)  # non D'Arcy flow
wellspec_dict['DZ']       = (0, wk_preferably_not, wk_okay, None,   True)   # non D'Arcy flow; use LENGTH value? or DZ

wellspec_dtype: Dict[str, Type] = { }  # mapping wellspec column key to expected data type

wellspec_dtype['IW']       = int
wellspec_dtype['JW']       = int
wellspec_dtype['L']        = int
wellspec_dtype['GRID']     = str
wellspec_dtype['RADW']     = float
wellspec_dtype['KHMULT']   = float
wellspec_dtype['STAT']     = str
wellspec_dtype['ANGLA']    = float
wellspec_dtype['ANGLV']    = float
wellspec_dtype['LENGTH']   = float
wellspec_dtype['KH']       = float
wellspec_dtype['SKIN']     = float
wellspec_dtype['PPERF']    = float
wellspec_dtype['ANGLE']    = float
wellspec_dtype['IRELPM']   = int
wellspec_dtype['RADB']     = float
wellspec_dtype['WI']       = float
wellspec_dtype['K']        = float
wellspec_dtype['LAYER']    = int
wellspec_dtype['DEPTH']    = float   # '#' causes nexus to use cell depth
wellspec_dtype['X']        = float   # use cell X for i.p. perf
wellspec_dtype['Y']        = float   # use cell Y for i.p. perf
wellspec_dtype['CELL']     = int     # CELL is for unstructured grids
wellspec_dtype['DTOP']     = float   # not compatible with ANGLA, ANGLV
wellspec_dtype['DBOT']     = float   # not compatible with ANGLA, ANGLV
wellspec_dtype['RADBP']    = float   # calculate as for RADB
wellspec_dtype['RADWP']    = float
wellspec_dtype['PORTYPE']  = str     # dual porosity: todo: need to check type
wellspec_dtype['FM']       = float
wellspec_dtype['MD']       = float
wellspec_dtype['PARENT']   = str
wellspec_dtype['MDCON']    = float
wellspec_dtype['SECT']     = str     # todo: need to check type
wellspec_dtype['FLOWSECT'] = str     # todo: need to check type
wellspec_dtype['ZONE']     = int
wellspec_dtype['GROUP']    = str
wellspec_dtype['TEMP']     = float
wellspec_dtype['IPTN']     = int
wellspec_dtype['D']        = float
wellspec_dtype['ND']       = str
wellspec_dtype['DZ']       = float
# yapf: enable


def increment_complaints(keyword):
   global wellspec_dict
   assert (keyword.upper() in wellspec_dict.keys())
   old_entry = wellspec_dict[keyword.upper()]
   wellspec_dict[keyword.upper()] = (old_entry[0] + 1, old_entry[1], old_entry[2], old_entry[3], old_entry[4])


def known_keyword(keyword):
   return keyword.upper() in wellspec_dict.keys()


def add_unknown_keyword(keyword):
   global wellspec_dict
   assert (not known_keyword(keyword))
   wellspec_dict[keyword.upper()] = (1, wk_unknown, wk_banned, None, False)  # assumes warning or error already given


def default_value(keyword):
   assert (known_keyword(keyword))
   return wellspec_dict[keyword][3]


def complaints(keyword):
   assert (known_keyword(keyword))
   return wellspec_dict[keyword][0]


def check_value(keyword, value):
   try:
      key = keyword.upper()
      if not known_keyword(key):
         return False
      if key in ['IW', 'JW', 'L', 'LAYER', 'IRELPM', 'CELL', 'SECT', 'FLOWSECT', 'ZONE', 'IPTN']:
         return int(value) > 0
      elif key == 'GRID':
         return len(str(value)) > 0
      elif key == 'STAT':
         return (str(value)).upper() in ['ON', 'OFF']
      elif key == 'ANGLA':
         return -360.0 <= float(value) and float(value) <= 360.0
      elif key == 'ANGLV':
         return 0.0 <= float(value) and float(value) <= 180.0
      elif key in ['RADW', 'RADB', 'RADWP', 'RADBP']:
         return float(value) > 0.0
      elif key in ['WI', 'LENGTH', 'KH', 'KHMULT', 'K', 'DZ']:
         return float(value) >= 0.0
      elif key == 'PPERF':
         return 0.0 <= float(value) and float(value) <= 1.0
      elif key == 'ANGLE':
         return 0.0 <= float(value) and float(value) <= 360.0
      elif key in ['SKIN', 'DEPTH', 'X', 'Y', 'TEMP']:
         float(value)
         return True
      else:
         return True
   except Exception:
      return False


def required_out_list():
   list = []
   for key in wellspec_dict.keys():
      if wellspec_dict[key][2] == wk_required:
         list.append(key)
   return list


def length_unit_conversion_applicable(keyword):
   assert (known_keyword(keyword))
   return wellspec_dict[keyword][4]


def load_wellspecs(wellspec_file, well = None, column_list = []):
   """Reads the Nexus wellspec file returning a dictionary of well name to pandas dataframe.

   args:
      wellspec_file (string): file path of ascii input file containing wellspec keywords
      well (optional, string): if present, only the data for the named well are loaded;
         if None, data for all wells are loaded
      column_list (list of strings, optional): if present, each dataframe returned contains
         these columns, in this order; if None, the resulting dictionary contains only
         well names as keys (each mapping to None rather than a dataframe); if an empty list,
         each dataframe contains the columns listed in the corresponding wellspec header, in
         the order found in the file

   returns:
      dictionary (string: pandas dataframe) mapping each well name found in the wellspec file
         to a dataframe containing the wellspec data
   """

   assert wellspec_file, 'no wellspec file specified'

   if column_list is not None:
      for column in column_list:
         assert column.upper() in wellspec_dict, 'unrecognized wellspec column name ' + str(column)
   selecting = bool(column_list)

   well_dict = {}  # maps from well name to pandas data frame with column_list as columns

   with open(wellspec_file, 'r') as fp:
      while True:
         found = kf.find_keyword(fp, 'WELLSPEC')
         if not found:
            break
         line = fp.readline()
         words = line.split()
         assert len(words) >= 2, 'missing well name after WELLSPEC keyword'
         well_name = words[1]
         if well and well_name.upper() != well.upper():
            continue
         if column_list is None:
            well_dict[well_name] = None
            continue
         kf.skip_blank_lines_and_comments(fp)
         line = kf.strip_trailing_comment(fp.readline()).upper()
         columns_present = line.split()
         if selecting:
            column_map = np.full((len(column_list),), -1, dtype = int)
            for i in range(len(column_list)):
               column = column_list[i].upper()
               if column in columns_present:
                  column_map[i] = columns_present.index(column)
            df = pd.DataFrame(columns = column_list)
         else:
            df = pd.DataFrame(columns = columns_present)
         while True:
            kf.skip_comments(fp)
            if kf.blank_line(fp):
               break  # unclear from Nexus doc what marks end of table
            if kf.specific_keyword_next(fp, 'WELLSPEC') or kf.specific_keyword_next(fp, 'WELLMOD'):
               break
            line = kf.strip_trailing_comment(fp.readline())
            words = line.split()
            assert len(words) >= len(columns_present), f'insufficient data in line of wellspec table {well} [{line}]'
            row_dict = {}
            if selecting:
               for col_index in range(len(column_list)):
                  column = column_list[col_index]
                  if column_map[col_index] < 0:
                     if column_list[col_index].upper() == 'GRID':
                        row_dict[column] = 'ROOT'
                     else:
                        row_dict[column] = np.NaN
                  else:
                     v = words[column_map[col_index]]
                     if v == 'NA':
                        row_dict[column] = np.NaN
                     elif v == '#':
                        row_dict[column] = v
                     else:
                        row_dict[column] = wellspec_dtype[column.upper()](v)
            else:
               for column, v in zip(columns_present, words[:len(columns_present)]):
                  if v == 'NA':
                     row_dict[column] = np.NaN
                  elif v == '#':
                     row_dict[column] = v
                  else:
                     row_dict[column] = wellspec_dtype[column](v)
            df = df.append(row_dict, ignore_index = True)

         if well:
            well_dict[well] = df
            break  # NB. if more than one table for a well, this function returns first, Nexus uses last
         well_dict[well_name] = df


#   log.debug(f'load-wellspecs returning:\n{well_dict}')

   return well_dict



import pytest
from resqpy import model


def test_empty_model():
   _ = model.Model()
   return


def test_all_imports():
   from resqpy import crs, derived_model, fault, grid, grid_surface, lines, organize
   from resqpy import property, rq_import, surface, time_series, well
   #    from resqpy.olio import *
   return



from pathlib import Path
from resqpy.organize import WellboreFeature

import pytest
import os
import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
from pandas.testing import assert_frame_equal

from resqpy.model import Model
from resqpy.crs import Crs
import resqpy.well
from resqpy.grid import RegularGrid
from resqpy.property import Property
import resqpy.olio.uuid as bu


def test_MdDatum(example_model_and_crs):

   # Set up a new datum
   model, crs = example_model_and_crs
   epc = model.epc_file
   data = dict(
      location = (0, -99999, 3.14),
      md_reference = 'mean low water',
   )
   datum = resqpy.well.MdDatum(parent_model = model, crs_uuid = crs.uuid, **data)
   uuid = datum.uuid

   # Save to disk and reload
   datum.create_xml()
   model.store_epc()

   del model, crs, datum
   model2 = Model(epc_file = epc)
   datum2 = resqpy.well.MdDatum(parent_model = model2, uuid = uuid)

   for key, expected_value in data.items():
      assert getattr(datum2, key) == expected_value, f"Issue with {key}"


def test_trajectory_iterators(example_model_with_well):
   # Test that trajectories can be found via iterators

   model, well_interp, datum, traj = example_model_with_well

   # Iterate via wells
   uuids_1 = []
   for well in model.iter_wellbore_interpretations():
      for traj in well.iter_trajectories():
         uuids_1.append(traj.uuid)

   # Iterate directly
   uuids_2 = []
   for traj in model.iter_trajectories():
      uuids_2.append(traj.uuid)

   # Two sets of trajectories should match, assuming all trajectories have a parent well
   assert len(uuids_1) > 0
   assert sorted(uuids_1) == sorted(uuids_2)


def test_logs(example_model_with_logs):

   model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

   # Check logs can be extracted from resqml dataset

   # Check exactly one wellbore frame exists in the model
   discovered_frames = 0
   for trajectory in model.iter_trajectories():
      for frame in trajectory.iter_wellbore_frames():
         discovered_frames += 1
   assert discovered_frames == 1

   # Check MDs
   mds = frame.node_mds
   assert isinstance(mds, np.ndarray)
   assert_array_almost_equal(mds, [1, 2, 3, 4])

   # Check logs
   log_list = list(log_collection.iter_logs())
   assert len(log_list) == 2

   # TODO: would be nice to write: log_collection.get_curve("GR")
   gr = log_list[0]
   assert gr.title == "GR"
   assert gr.uom == "gAPI"
   assert_array_almost_equal(gr.values(), [1, 2, 1, 2])

   nphi = log_list[1]
   assert nphi.title == 'NPHI'

   # TODO: get more units working
   # assert nphi.uom == "v/v"
   assert_array_almost_equal(nphi.values(), [0.1, 0.1, np.NaN, np.NaN])


def test_logs_conversion(example_model_with_logs):

   model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

   # Pandas
   df = log_collection.to_df()
   df_expected = pd.DataFrame(data = {"GR": [1, 2, 1, 2], "NPHI": [0.1, 0.1, np.NaN, np.NaN]}, index = [1, 2, 3, 4])
   assert_frame_equal(df_expected, df, check_dtype = False)

   # LAS
   las = log_collection.to_las()
   assert las.well.WELL.value == 'well A'

   gr = las.get_curve("GR")
   assert gr.unit.casefold() == "GAPI".casefold()
   assert_array_almost_equal(gr.data, [1, 2, 1, 2])

   nphi = las.get_curve("NPHI")
   # assert nphi.unit == "GAPI"
   assert_array_almost_equal(nphi.data, [0.1, 0.1, np.NaN, np.NaN])


# Trajectory


def test_Trajectory_add_well_feature_and_interp(example_model_and_crs):

   # Prepare an example Trajectory without a well feature
   wellname = "Hullabaloo"
   model, crs = example_model_and_crs
   datum = resqpy.well.MdDatum(parent_model = model,
                               crs_uuid = crs.uuid,
                               location = (0, 0, -100),
                               md_reference = 'kelly bushing')
   datum.create_xml()
   traj = resqpy.well.Trajectory(parent_model = model, md_datum = datum, well_name = wellname)

   # Add the well interp
   assert traj.wellbore_feature is None
   assert traj.wellbore_interpretation is None
   traj.create_feature_and_interpretation()

   # Check well is present
   assert traj.wellbore_feature is not None
   assert traj.wellbore_feature.feature_name == wellname


# Deviation Survey tests


def test_DeviationSurvey(example_model_with_well):
   # Test that all attrbutes are correctly saved and loaded from disk

   # --------- Arrange ----------
   # Create a Deviation Survey object in memory

   # Load example model from a fixture
   model, well_interp, datum, traj = example_model_with_well
   epc_path = model.epc_file

   # Create the survey
   data = dict(
      title = 'Majestic Umlaut ö',
      originator = 'Thor, god of sparkles',
      md_uom = 'ft',
      angle_uom = 'rad',
      is_final = True,
   )
   array_data = dict(
      measured_depths = np.array([1, 2, 3], dtype = float),
      azimuths = np.array([4, 5, 6], dtype = float),
      inclinations = np.array([7, 8, 9], dtype = float),
      first_station = np.array([0, -1, 999], dtype = float),
   )
   survey = resqpy.well.DeviationSurvey(
      parent_model = model,
      represented_interp = well_interp,
      md_datum = datum,
      **data,
      **array_data,
   )
   survey_uuid = survey.uuid

   # ----------- Act ---------

   # Save to disk
   survey.write_hdf5()
   survey.create_xml()
   model.store_epc()
   model.h5_release()

   # import shutil
   # shutil.copy(epc_path, r'C:\Temp')
   # shutil.copy(epc_path.replace('.epc', '.h5'), r'C:\Temp')

   # Clear memory
   del model, well_interp, datum, traj, survey

   # Reload from disk
   model2 = Model(epc_file = epc_path)
   survey2 = resqpy.well.DeviationSurvey(model2, uuid = survey_uuid)

   # --------- Assert --------------
   # Check all attributes were loaded from disk correctly

   for key, expected_value in data.items():
      assert getattr(survey2, key) == expected_value, f"Error for {key}"

   for key, expected_value in array_data.items():
      assert_array_almost_equal(getattr(survey2, key), expected_value, err_msg = f"Error for {key}")


# BlockedWell


def test_wellspec_properties(example_model_and_crs):
   model, crs = example_model_and_crs
   grid = RegularGrid(model,
                      extent_kji = (3, 4, 3),
                      dxyz = (50.0, -50.0, 50.0),
                      origin = (0.0, 0.0, 100.0),
                      crs_uuid = crs.uuid,
                      set_points_cached = True)
   grid.write_hdf5()
   grid.create_xml(write_geometry = True)
   wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
   well_name = 'DOGLEG'
   source_df = pd.DataFrame([[2, 2, 1, 0.0, 0.0, 0.0, 0.25], [2, 2, 2, 0.45, -90.0, 2.5, 0.25],
                             [2, 3, 2, 0.45, -90.0, 1.0, 0.20], [2, 3, 3, 0.0, 0.0, -0.5, 0.20]],
                            columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW'])
   with open(wellspec_file, 'w') as fp:
      fp.write(F'WELLSPEC {well_name}\n')
      for col in source_df.columns:
         fp.write(f' {col:>6s}')
      fp.write('\n')
      for row_index in range(len(source_df)):
         row = source_df.iloc[row_index]
         for col in source_df.columns:
            if col in ['IW', 'JW', 'L']:
               fp.write(f' {int(row[col]):6d}')
            else:
               fp.write(f' {row[col]:6.2f}')
         fp.write('\n')
   bw = resqpy.well.BlockedWell(model,
                                wellspec_file = wellspec_file,
                                well_name = well_name,
                                use_face_centres = True,
                                add_wellspec_properties = True)
   assert bw is not None
   bw_uuid = bw.uuid
   skin_uuid = model.uuid(title = 'SKIN', related_uuid = bw.uuid)
   assert skin_uuid is not None
   skin_prop = Property(model, uuid = skin_uuid)
   assert skin_prop is not None
   assert_array_almost_equal(skin_prop.array_ref(), [0.0, 2.5, 1.0, -0.5])
   model.store_epc()
   # re-open model from persistent storage
   model = Model(model.epc_file)
   bw2_uuid = model.uuid(obj_type = 'BlockedWellboreRepresentation', title = 'DOGLEG')
   assert bw2_uuid is not None
   bw2 = resqpy.well.BlockedWell(model, uuid = bw2_uuid)
   assert bu.matching_uuids(bw_uuid, bw2_uuid)
   df2 = bw.dataframe(extra_columns_list = ['ANGLV', 'ANGLA', 'LENGTH', 'SKIN', 'RADW'], use_properties = True)
   assert df2 is not None
   assert len(df2.columns) == 8
   for col in ['ANGLV', 'ANGLA', 'SKIN', 'RADW']:
      assert_array_almost_equal(np.array(source_df[col]), np.array(df2[col]))
   df3 = bw.dataframe(extra_columns_list = ['ANGLV', 'ANGLA', 'LENGTH', 'SKIN', 'RADW'],
                      use_properties = ['SKIN', 'RADW'])
   for col in ['SKIN', 'RADW']:
      assert_array_almost_equal(np.array(source_df[col]), np.array(df3[col]))
